/*
 * LAMIP-0.0.3 - DSP Channel Mapper plugin
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <plug_dsp.h>

/* channel mapper */
struct {
  lamip_channel_mapper	map;
  lamipPCMFormat	fmt;		/* PCMFormat of input */
  int			buf_ratio;	/* if > 1, malloc() */
} ch_map;


static int	chanmap_init(const lamipPluginHandle*);
static void	chanmap_cleanup(void);
static int	chanmap_setup(lamipPCMInfo*);
static int	chanmap_probe(DSPMaterial*);
static int	chanmap_filter(DSPMaterial*);

static DSPPlugin	chanmap;
static int		channels;

#ifdef DEBUGING
 #define DEBUG(a) fprintf(stderr, "DEBUG_CHANMAP: %s\n", a)
#else
 #define DEBUG(a)
#endif


/****************************************************************************/

DSPPlugin* lamip_dsp_info(void)
{
  chanmap.common.name		= "chanmap";
  chanmap.common.description	= "chanmap - Mixes the stream to the number of output channels";
  chanmap.common.init		= chanmap_init;
  chanmap.common.cleanup	= chanmap_cleanup;
  chanmap.setup			= chanmap_setup;
  chanmap.probe			= chanmap_probe;
  chanmap.filter		= chanmap_filter;

  return(&chanmap);
}

static int chanmap_init(const lamipPluginHandle* conf)
{
  lamipPluginHandle*	config;
  int			val_channels;

  DEBUG("chanmap_init");
  config = (lamipPluginHandle*)conf;
  if(!lamip_cfg_getExist(conf, "channels")){
    lamip_send_message("CHAN_MAP: chanmap_init: no config value for \"channels\" found, resetting to default...\n");
    val_channels = 2;
    lamip_cfg_setInt(config, "channels", val_channels);
  }else{
    val_channels = lamip_cfg_getInt(conf, "channels");
  }
  channels = val_channels;
  ch_map.buf_ratio = 1;
  return(1);
}

static void	chanmap_cleanup(void)
{
  DEBUG("chanmap_cleanup");
}

static int chanmap_setup(lamipPCMInfo* setup)
{
  DEBUG("chanmap_setup");
  /* We always want "channels" number of channels */
  setup->channels = channels;
  return(1);
}

static int chanmap_probe(DSPMaterial* material)
{
  DEBUG("chanmap_probe");
  /* Nothing should go wrong... Just set our number of channels */
  material->output_pcm_info = material->input_pcm_info;
  material->output_pcm_info.channels = channels;

  if((channels>0) && (material->input_pcm_info.channels != channels)){
    ch_map.map = lamip_pcm_getChannelMapper(material->input_pcm_info.channels, channels);
    if(ch_map.map){
      lamip_send_message("CHAN_MAP: channel mapping %lu => %lu", material->input_pcm_info.channels, channels);
      ch_map.buf_ratio = channels / material->input_pcm_info.channels;
      if(ch_map.buf_ratio * material->input_pcm_info.channels < channels)
	ch_map.buf_ratio++;
      material->output_pcm_info.channels = channels;
      return(1);
    }else{
      lamip_send_message("CHAN_MAP: channel mapping %lu => %lu not available", material->input_pcm_info.channels, channels);
      return(0);
    }
  }
  lamip_send_message("CHAN_MAP: no change on pcm flow");
  return(1);
}

static int chanmap_filter(DSPMaterial* material)
{
  DEBUG("chanmap_filter");
  if(ch_map.buf_ratio > 1){
    material->output_buffer_len = material->input_buffer_len * ch_map.buf_ratio;
    material->output_buffer = malloc(material->output_buffer_len);
    if(!material->output_buffer) {
      fprintf(stderr, "CHAN_MAP: malloc error!");
      return(0);
    }
    memset(material->output_buffer, 0, material->output_buffer_len);
    if(ch_map.map){ /* do the mapping */
      material->output_buffer_len = ch_map.map(material->output_buffer, material->output_buffer_len, material->input_buffer, material->input_buffer_len, material->input_pcm_info.format);
    }else{
      lamip_send_message("CHAN_MAP: didn't get correct ch_map.map function !");
      free(material->output_buffer);
      material->output_buffer = 0;
    }
  }

  if(ch_map.buf_ratio==1 && !material->output_buffer){
    lamip_send_message("CHAN_MAP: didn't do the the channel mapping !");
    material->output_buffer_len = material->input_buffer_len;
    material->output_buffer = malloc(material->output_buffer_len);
    if(!material->output_buffer) {
      fprintf(stderr, "CHAN_MAP: malloc errror!\n");
      return(0);
    }
    memcpy(material->output_buffer, material->input_buffer, material->output_buffer_len);
  }
  return(1);
}
