/*
 * LAMIP - Linux Audio Multiple Interface Player
 * copyright (c) 2003 - 2004 Frederic Fondriest
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licensse as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 ******************************************************************************
 *
 * There are three non-static functions here:
 *
 * lamip_get_channel_mapper(const unsigned long, const unsigned long);
 * lamip_get_resampler(const unsigned long, const unsigned long);
 * lamip_get_format_changer(const lamipPCMFormat, const lamipPCMFormat);
 *
 * , and they return function pointer to suitable conversion function
 *
*/

#include <stdio.h>
#include <stdlib.h>

#include "plug_dsp.h"

/* portable? hum only for 32bits prossesors */
#include <stdint.h>
/*
typedef signed char		int8_t;
typedef short int		int16_t;
typedef int			int32_t;
typedef unsigned char		uint8_t;
typedef unsigned short int	uint16_t;
typedef unsigned int		uint32_t;
*/

#define UINT16_SWAP(val)	((uint16_t) ((val >> 8) | (val << 8)))
#define UINT32_SWAP(val)	((uint32_t) ((val >> 24) | ((val & 0x00ff0000) >> 8) | ((val & 0x0000ff00) << 8) | (val << 24)))

#ifdef WORDS_BIGENDIAN

 #define IS_BIG_ENDIAN 1
 #define INT16_TO_LE(val)	((int16_t) UINT16_SWAP (val))
 #define UINT16_TO_LE(val)	(UINT16_SWAP (val))
 #define INT32_TO_LE(val)	((int32_t) UINT32_SWAP (val))
 #define UINT32_TO_LE(val)	(UINT32_SWAP (val))
 #define INT16_TO_BE(val)	((int16_t) (val))
 #define UINT16_TO_BE(val)	((uint16_t) (val))
 #define INT32_TO_BE(val)	((int32_t) (val))
 #define UINT32_TO_BE(val)	((uint32_t) (val))

#else /* little endian */

 #define IS_BIG_ENDIAN 0
 #define INT16_TO_LE(val)	((int16_t) (val))
 #define UINT16_TO_LE(val)	((uint16_t) (val))
 #define INT32_TO_LE(val)	((int32_t) (val))
 #define UINT32_TO_LE(val)	((uint32_t) (val))
 #define INT16_TO_BE(val)	((int16_t) UINT16_SWAP (val))
 #define UINT16_TO_BE(val)	(UINT16_SWAP (val))
 #define INT32_TO_BE(val)	((int32_t) UINT32_SWAP (val))
 #define UINT32_TO_BE(val)	(UINT32_SWAP (val))

#endif

/* they are symmetric */
#define INT16_FROM_LE(val)     (INT16_TO_LE (val))
#define UINT16_FROM_LE(val)    (UINT16_TO_LE (val))
#define INT32_FROM_LE(val)     (INT32_TO_LE (val))
#define UINT32_FROM_LE(val)    (UINT32_TO_LE (val))

#define INT16_FROM_BE(val)     (INT16_TO_BE (val))
#define UINT16_FROM_BE(val)    (UINT16_TO_BE (val))
#define INT32_FROM_BE(val)     (INT32_TO_BE (val))
#define UINT32_FROM_BE(val)    (UINT32_TO_BE (val))

#define MIN(a, b) ((a) < (b) ? (a) : (b))


static size_t dummy_with_fmt(void*, const size_t,
			     const void*, const size_t, const lamipPCMFormat);
static size_t dummy(void*, const size_t, const void*, const size_t);
static size_t fmt_swap_endian(void*, const size_t, const void*, const size_t);
static size_t fmt_swap_sign16(void*, const size_t, const void*, const size_t);
static size_t fmt_swap_sign8(void *, const size_t, const void*, const size_t);
static size_t fmt_swap_sign_and_endian_to_native(void*, const size_t,
						 const void*, const size_t);
static size_t fmt_swap_sign_and_endian_to_alien(void*, const size_t,
						const void*, const size_t);
static size_t fmt_to_8_native_endian(void*, const size_t,
				     const void*, const size_t);
static size_t fmt_to_8_native_endian_swap_sign(void*, const size_t,
					       const void*, const size_t);
static size_t fmt_to_8_alien_endian(void*, const size_t,
				    const void*, const size_t);
static size_t fmt_to_8_alien_endian_swap_sign(void*, const size_t,
					      const void*, const size_t);
static size_t fmt_to_16_native_endian(void*, const size_t,
				      const void*, const size_t);
static size_t fmt_to_16_native_endian_swap_sign(void*, const size_t,
						const void*, const size_t);
static size_t fmt_to_16_alien_endian(void*, const size_t,
				     const void*, const size_t);
static size_t fmt_to_16_alien_endian_swap_sign(void*, const size_t,
					       const void*, const size_t);
static size_t map_1_to_2(void*, const size_t,
			 const void*, const size_t, lamipPCMFormat);
//static size_t map_1_to_4(void*, const size_t,
//			 const void*, const size_t, lamipPCMFormat);
//static size_t map_1_to_6(void*, const size_t,
//			 const void*, const size_t, lamipPCMFormat);
static size_t map_2_to_1(void*, const size_t,
			 const void*, const size_t, lamipPCMFormat);
static size_t map_2_to_4(void*, const size_t,
			 const void*, const size_t, lamipPCMFormat);
static size_t map_2_to_6(void*, const size_t,
			 const void*, const size_t, lamipPCMFormat);
//static size_t map_4_to_2(void*, const size_t,
//			 const void*, const size_t, lamipPCMFormat);
//static size_t map_4_to_6(void*, const size_t,
//			 const void*, const size_t, lamipPCMFormat);
static size_t map_6_to_4(void*, const size_t,
			 const void*, const size_t, lamipPCMFormat);
static size_t map_6_to_2(void*, const size_t,
			 const void*, const size_t, lamipPCMFormat);


void lamip_pcm_formatDecomp(lamipPCMFormat fmt, unsigned int *samplesize, int *bigendian, int *sign)
{
  switch(fmt){
  case PCM_FORMAT_S8:
    if(samplesize)
      *samplesize = 1;
    if(bigendian)
      *bigendian = 0;
    if(sign)
      *sign = 1;
    break;
  case PCM_FORMAT_U8:
    if(samplesize)
      *samplesize = 1;
    if(bigendian)
      *bigendian = 0;
    if(sign)
      *sign = 0;
    break;
  case PCM_FORMAT_S16_LE:
    if(samplesize)
      *samplesize = 2;
    if(bigendian)
      *bigendian = 0;
    if(sign)
      *sign = 1;
    break;
  case PCM_FORMAT_S16_BE:
    if(samplesize)
      *samplesize = 2;
    if(bigendian)
      *bigendian = 1;
    if(sign)
      *sign = 1;
    break;
  case PCM_FORMAT_U16_LE:
    if(samplesize)
      *samplesize = 2;
    if(bigendian)
      *bigendian = 0;
    if(sign)
      *sign = 0;
    break;
  case PCM_FORMAT_U16_BE:
    if(samplesize)
      *samplesize = 2;
    if(bigendian)
      *bigendian = 1;
    if(sign)
      *sign = 0;
    break;
  case PCM_FORMAT_S24_LE:
    if(samplesize)
      *samplesize = 4; /* only first 3 bytes used */
    if(bigendian)
      *bigendian = 0;
    if(sign)
      *sign = 1;
    break;
  case PCM_FORMAT_S24_BE:
    if(samplesize)
      *samplesize = 4;
    if(bigendian)
      *bigendian = 1;
    if(sign)
      *sign = 1;
    break;
  case PCM_FORMAT_U24_LE:
    if(samplesize)
      *samplesize = 4;
    if(bigendian)
      *bigendian = 0;
    if(sign)
      *sign = 0;
    break;
  case PCM_FORMAT_U24_BE:
    if(samplesize)
      *samplesize = 4;
    if(bigendian)
      *bigendian = 1;
    if(sign)
      *sign = 0;
    break;
  case PCM_FORMAT_S32_LE:
    if(samplesize)
      *samplesize = 4;
    if(bigendian)
      *bigendian = 0;
    if(sign)
      *sign = 1;
    break;
  case PCM_FORMAT_S32_BE:
    if(samplesize)
      *samplesize = 4;
    if(bigendian)
      *bigendian = 1;
    if(sign)
      *sign = 1;
    break;
  case PCM_FORMAT_U32_LE:
    if(samplesize)
      *samplesize = 4;
    if(bigendian)
      *bigendian = 0;
    if(sign)
      *sign = 0;
    break;
  case PCM_FORMAT_U32_BE:
    if(samplesize)
      *samplesize = 4;
    if(bigendian)
      *bigendian = 1;
    if(sign)
      *sign = 0;
  }
}

/* from xmms */
void lamip_pcm_to_s16ne(lamipPCMFormat fmt, void *data, void *left, void *right, unsigned int nch, int max)
{
  int16_t*	ptr16;
  uint16_t*	ptru16;
  uint8_t*	ptru8;
  int16_t*	ptrl = left, *ptrr = right;
  int		i;

  switch(fmt){
  case PCM_FORMAT_U8:
    ptru8 = data;
    if(nch == 1)
      for(i = 0; i < max; i++)
	ptrl[i] = ((*ptru8++) ^ 128) << 8;
    else
      for(i = 0; i < max; i++){
	ptrl[i] = ((*ptru8++) ^ 128) << 8;
	ptrr[i] = ((*ptru8++) ^ 128) << 8;
      }
    break;
  case PCM_FORMAT_S8:
    ptru8 = data;
    if(nch == 1)
      for(i = 0; i < max; i++)
	ptrl[i] = (*ptru8++) << 8;
    else
      for(i = 0; i < max; i++){
	ptrl[i] = (*ptru8++) << 8;
	ptrr[i] = (*ptru8++) << 8;
      }
    break;
  case PCM_FORMAT_U16_LE:
    ptru16 = data;
    if(nch == 1)
      for(i = 0; i < max; i++, ptru16++)
	ptrl[i] = UINT16_FROM_LE(*ptru16) ^ 32768;
    else
      for(i = 0; i < max; i++){
	ptrl[i] = UINT16_FROM_LE(*ptru16) ^ 32768;
	ptru16++;
	ptrr[i] = UINT16_FROM_LE(*ptru16) ^ 32768;
	ptru16++;
      }
    break;
  case PCM_FORMAT_U16_BE:
    ptru16 = data;
    if(nch == 1)
      for(i = 0; i < max; i++, ptru16++)
	ptrl[i] = UINT16_FROM_BE(*ptru16) ^ 32768;
    else
      for(i = 0; i < max; i++){
	ptrl[i] = UINT16_FROM_BE(*ptru16) ^ 32768;
	ptru16++;
	ptrr[i] = UINT16_FROM_BE(*ptru16) ^ 32768;
	ptru16++;
      }
    break;
    /*
      case PCM_FORMAT_U16_NE:
      ptru16 = data;
      if (nch == 1)
      for (i = 0; i < max; i++)
      ptrl[i] = (*ptru16++) ^ 32768;
      else
      for (i = 0; i < max; i++)
      {
      ptrl[i] = (*ptru16++) ^ 32768;
      ptrr[i] = (*ptru16++) ^ 32768;
      }
      break;
    */
  case PCM_FORMAT_S16_LE:
    ptr16 = data;
    if(nch == 1)
      for(i = 0; i < max; i++, ptr16++)
	ptrl[i] = INT16_FROM_LE(*ptr16);
    else
      for(i = 0; i < max; i++){
	ptrl[i] = INT16_FROM_LE(*ptr16);
	ptr16++;
	ptrr[i] = INT16_FROM_LE(*ptr16);
	ptr16++;
      }
    break;
  case PCM_FORMAT_S16_BE:
    ptr16 = data;
    if(nch == 1)
      for(i = 0; i < max; i++, ptr16++)
	ptrl[i] = INT16_FROM_BE(*ptr16);
    else
      for(i = 0; i < max; i++){
	ptrl[i] = INT16_FROM_BE(*ptr16);
	ptr16++;
	ptrr[i] = INT16_FROM_BE(*ptr16);
	ptr16++;
      }
    break;
    /*
      case PCM_FORMAT_S16_NE:
      ptr16 = data;
      if (nch == 1)
      for (i = 0; i < max; i++)
      ptrl[i] = (*ptr16++);
      else
      for (i = 0; i < max; i++)
      {
      ptrl[i] = (*ptr16++);
      ptrr[i] = (*ptr16++);
      }
      break;
    */
  default:
    fprintf(stderr, "unsupported\n");
    break;
  }
}

lamip_channel_mapper lamip_pcm_getChannelMapper(const unsigned long old_nch, const unsigned long new_nch)
{
  if(old_nch == new_nch)
    return(dummy_with_fmt);
  if(old_nch == 1 && new_nch == 2)
    return(map_1_to_2);
  if(old_nch == 2 && new_nch == 1)
    return(map_2_to_1);
  if(old_nch == 2 && new_nch == 4)
    return(map_2_to_4);
  if(old_nch == 2 && new_nch == 6)
    return(map_2_to_6);
  if(old_nch == 6 && new_nch == 2)
    return(map_6_to_2);
  if(old_nch == 6 && new_nch == 4)
    return(map_6_to_4);
  return(NULL);
}

lamip_resampler lamip_pcm_getResampler(const unsigned long old_sr, const unsigned long new_sr)
{
  if(old_sr == new_sr)
    return(dummy_with_fmt);
  return(NULL);
}

static size_t dummy_with_fmt(void *buf, const size_t buf_len, const void *data, const size_t data_len, const lamipPCMFormat fmt)
{
  return(data_len);
}

static size_t dummy(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  return(data_len);
}

/*
 * based on xmms oss output plugin
*/
static size_t	fmt_swap_endian(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  uint16_t*	  bp = buf;
  const uint16_t* dp = data;
  int		  i;

  for(i=0; i<data_len; i+=2, dp++)
    *bp++ = UINT16_SWAP(*dp);
  return(i);
}

/*
 * signed <=> unsigned for 16-bit data
*/
static size_t fmt_swap_sign16(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  int16_t*	 bp = buf;
  const int16_t* dp = data;
  int		 i;

  for(i=0; i<data_len; i+=2)
    *bp++ = *dp++ ^ (1 << 15);
  return(i);
}

/*
 * signed <=> unsigned for 8-bit data
*/
static size_t fmt_swap_sign8(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  int8_t*	bp = buf;
  const int8_t* dp = data;
  int		i;

  for(i=0; i<data_len; i++)
    *bp++ = *dp++ ^ (1 << 7);
  return(i);
}

/*
 * le machine:
 * signed be => unsigned le
 * unsigned be => signed le
 *
 * be machine:
 * signed le => unsigned be
 * unsigned le => signed be
 */
static size_t fmt_swap_sign_and_endian_to_native(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  uint16_t*	  bp = buf;
  const uint16_t* dp = data;
  int		  i;

  for(i=0; i<data_len; i+=2, dp++)
    *bp++ = UINT16_SWAP(*dp) ^ (1 << 15);
  return(i);
}

/*
 * le machine:
 * signed le => unsigned be
 * unsigned le => signed be
 *
 * be machine:
 * signed be => unsigned le
 * unsigned be => signed le
*/
static size_t fmt_swap_sign_and_endian_to_alien(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  uint16_t*	  bp = buf;
  const uint16_t* dp = data;
  int		  i;

  for(i=0; i<data_len; i+=2)
    *bp++ = UINT16_SWAP(*dp ^ (1 << 15));
  return(i);
}


/*
 * round to 8-bit for machine having same endian as data
*/
static size_t fmt_to_8_native_endian(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  int8_t*	 bp = buf;
  const int16_t* dp = data;
  int		 i;

  for(i=0; i<data_len/2; i++)
    *bp++ = (*dp++ + (int16_t) 0x80) >> 8;
  return(i);
}

/*
 * and signed <=> unsigned
*/
static size_t fmt_to_8_native_endian_swap_sign(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  int8_t*	 bp = buf;
  const int16_t* dp = data;
  int		 i;

  for(i=0; i<data_len/2; i++)
    *bp++ = ((*dp++ + (int16_t) 0x80) >> 8) ^ (1 << 7);
  return(i);
}


/*
 * round to 8-bit for machine having different endian from data
*/
static size_t fmt_to_8_alien_endian(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  int8_t*	 bp = buf;
  const int16_t* dp = data;
  int		 i;

  for(i=0; i<data_len/2; i++, dp++)
    *bp++ = (((int16_t) UINT16_SWAP(*dp)) + (int16_t) 0x80) >> 8;
  return(i);
}

/*
 * and signed <=> unsigned
*/
static size_t fmt_to_8_alien_endian_swap_sign(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  int8_t*	 bp = buf;
  const int16_t* dp = data;
  int		 i;

  for(i=0; i<data_len/2; i++)
    *bp++ = ((((int16_t) UINT16_SWAP(*dp)) + (int16_t) 0x80) >> 8) ^ (1 << 7);
  return(i);
}

/*
 * generate 16-bit data with machines's endian
*/
static size_t fmt_to_16_native_endian(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  uint16_t*	 bp = buf;
  const uint8_t* dp = data;
  int		 i;

  for(i=0; i<data_len/2; i++)
    *bp++ = *dp++ << 8;
  return(i*2);
}

/*
 * and signed <=> unsigned
*/
static size_t fmt_to_16_native_endian_swap_sign(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  uint16_t*	 bp = buf;
  const uint8_t* dp = data;
  int		 i;

  for(i=0; i<data_len/2; i++)
    *bp++ = (*dp++ << 8) ^ (1 << 15);
  return(i*2);
}


/*
 * generate 16-bit data with endian different from machines's
*/
static size_t fmt_to_16_alien_endian(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  uint16_t*	 bp = buf;
  const uint8_t* dp = data;
  int		 i;

  for(i=0; i<data_len/2; i++)
    *bp++ = *dp++;
  return(i*2);
}

/*
 * and signed <=> unsigned
 */
static size_t fmt_to_16_alien_endian_swap_sign(void *buf, const size_t buf_len, const void *data, const size_t data_len)
{
  uint16_t*	 bp = buf;
  const uint8_t* dp = data;
  int		 i;

  for(i=0; i<data_len/2; i++)
    *bp++ = *dp++ ^ (1 << 7);
  return(i*2);
}

lamip_format_changer lamip_pcm_getFormatChanger(const lamipPCMFormat old_fmt, const lamipPCMFormat new_fmt)
{
  if(old_fmt == new_fmt)
    return(dummy);

  if((new_fmt == PCM_FORMAT_U16_BE && old_fmt == PCM_FORMAT_U16_LE) ||
     (new_fmt == PCM_FORMAT_U16_LE && old_fmt == PCM_FORMAT_U16_BE) ||
     (new_fmt == PCM_FORMAT_S16_BE && old_fmt == PCM_FORMAT_S16_LE) ||
     (new_fmt == PCM_FORMAT_S16_LE && old_fmt == PCM_FORMAT_S16_BE))
    return(fmt_swap_endian);
  if((new_fmt == PCM_FORMAT_U16_BE && old_fmt == PCM_FORMAT_S16_BE) ||
     (new_fmt == PCM_FORMAT_U16_LE && old_fmt == PCM_FORMAT_S16_LE) ||
     (new_fmt == PCM_FORMAT_S16_BE && old_fmt == PCM_FORMAT_U16_BE) ||
     (new_fmt == PCM_FORMAT_S16_LE && old_fmt == PCM_FORMAT_U16_LE))
    return(fmt_swap_sign16);

  if((IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_BE && old_fmt == PCM_FORMAT_S16_LE) ||
       (new_fmt == PCM_FORMAT_S16_BE && old_fmt == PCM_FORMAT_U16_LE))) ||
     (!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_LE && old_fmt == PCM_FORMAT_S16_BE) ||
       (new_fmt == PCM_FORMAT_S16_LE && old_fmt == PCM_FORMAT_U16_BE))))
    return(fmt_swap_sign_and_endian_to_native);
  if((!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_BE && old_fmt == PCM_FORMAT_S16_LE) ||
       (new_fmt == PCM_FORMAT_S16_BE && old_fmt == PCM_FORMAT_U16_LE))) ||
     (IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_LE && old_fmt == PCM_FORMAT_S16_BE) ||
       (new_fmt == PCM_FORMAT_S16_LE && old_fmt == PCM_FORMAT_U16_BE))))
    return(fmt_swap_sign_and_endian_to_alien);

  if((IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_U16_BE) ||
       (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_S16_BE))) ||
     (!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_U16_LE) ||
       (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_S16_LE))))
    return(fmt_to_8_native_endian);

  if((IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_S16_BE) ||
       (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_U16_BE))) ||
     (!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_S16_LE) ||
       (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_U16_LE))))
    return(fmt_to_8_native_endian_swap_sign);
  if((!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_U16_BE) ||
       (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_S16_BE))) ||
     (IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_U16_LE) ||
       (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_S16_LE))))
    return(fmt_to_8_alien_endian);
  if((!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_S16_BE) ||
       (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_U16_BE))) ||
     (IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_S16_LE) ||
       (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_U16_LE))))
    return(fmt_to_8_alien_endian_swap_sign);
  if((new_fmt == PCM_FORMAT_U8 && old_fmt == PCM_FORMAT_S8) ||
     (new_fmt == PCM_FORMAT_S8 && old_fmt == PCM_FORMAT_U8))
    return(fmt_swap_sign8);
  if((IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_BE && old_fmt == PCM_FORMAT_U8) ||
       (new_fmt == PCM_FORMAT_S16_BE && old_fmt == PCM_FORMAT_S8))) ||
     (!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_LE && old_fmt == PCM_FORMAT_U8) ||
       (new_fmt == PCM_FORMAT_S16_LE && old_fmt == PCM_FORMAT_S8))))
    return(fmt_to_16_native_endian);
  if((IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_BE && old_fmt == PCM_FORMAT_S8) ||
       (new_fmt == PCM_FORMAT_S16_BE && old_fmt == PCM_FORMAT_U8))) ||
     (!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_LE && old_fmt == PCM_FORMAT_S8) ||
       (new_fmt == PCM_FORMAT_S16_LE && old_fmt == PCM_FORMAT_U8))))
    return(fmt_to_16_native_endian_swap_sign);
  if((!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_BE && old_fmt == PCM_FORMAT_U8) ||
       (new_fmt == PCM_FORMAT_S16_BE && old_fmt == PCM_FORMAT_S8))) ||
     (IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_LE && old_fmt == PCM_FORMAT_U8) ||
       (new_fmt == PCM_FORMAT_S16_LE && old_fmt == PCM_FORMAT_S8))))
    return(fmt_to_16_alien_endian);
  if((!IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_BE && old_fmt == PCM_FORMAT_S8) ||
       (new_fmt == PCM_FORMAT_S16_BE && old_fmt == PCM_FORMAT_U8))) ||
     (IS_BIG_ENDIAN &&
      ((new_fmt == PCM_FORMAT_U16_LE && old_fmt == PCM_FORMAT_S8) ||
       (new_fmt == PCM_FORMAT_S16_LE && old_fmt == PCM_FORMAT_U8))))
    return(fmt_to_16_alien_endian_swap_sign);
  fprintf(stderr, "No format conversion available");
  return(NULL);
}

static size_t map_1_to_2(void *buf, const size_t buf_len, const void *data, const size_t data_len, lamipPCMFormat fmt)
{
  int		samples;
  int		samplesize;
  size_t	out_len;

  lamip_pcm_formatDecomp(fmt, &samplesize, 0, 0);
  samples = data_len/(1*samplesize);
  out_len = data_len*2;

  switch(fmt){
  case PCM_FORMAT_U8:
  case PCM_FORMAT_S8:
    {
      uint8_t*		output = buf;
      const uint8_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *input++;
      }
      break;
    }
  case PCM_FORMAT_S16_LE:
  case PCM_FORMAT_S16_BE:
  case PCM_FORMAT_U16_LE:
  case PCM_FORMAT_U16_BE:
    {
      uint16_t*		output = buf;
      const uint16_t*	input = data;
      
      while(samples--){
	(*output++) = (*input);
	(*output++) = (*input++);
      }
      
      break;
    }
  case PCM_FORMAT_S24_LE:
  case PCM_FORMAT_S24_BE:
  case PCM_FORMAT_U24_LE:
  case PCM_FORMAT_U24_BE:
    {
      uint32_t*		output = buf;
      const uint32_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *input++;
      }
      break;
    }
  case PCM_FORMAT_S32_LE:
  case PCM_FORMAT_S32_BE:
  case PCM_FORMAT_U32_LE:
  case PCM_FORMAT_U32_BE:
    {
      uint32_t*		output = buf;
      const uint32_t*	input = data;

      while (samples--) {
	*output++ = *input;
	*output++ = *input++;
      }
      break;
    }
  default:
    {
      fprintf(stderr, "map_2_to_1: unknown format\n");
      return(0);
    }
  }
  return(out_len);
}

static size_t map_2_to_1(void *buf, const size_t buf_len, const void *data, const size_t data_len, lamipPCMFormat fmt)
{
  int i, samples, samplesize;
  size_t out_len;

  lamip_pcm_formatDecomp(fmt, &samplesize, NULL, NULL);
  samples = MIN(data_len / (2 * samplesize), buf_len / (1 * samplesize));
  out_len = samples * (1 * samplesize);

  switch(fmt){
  case PCM_FORMAT_U8:
    {
      uint8_t*		output = buf;
      const uint8_t*	input = data;

      for(i=0; i<samples; i++){
	uint16_t tmp;

	tmp = *input++;
	tmp += *input++;
	*output++ = tmp / 2;
      }
      break;
    }
  case PCM_FORMAT_S8:
    {
      int8_t*		output = buf;
      const int8_t*	input = data;

      for(i=0; i<samples; i++){
	int16_t tmp;

	tmp = *input++;
	tmp += *input++;
	*output++ = tmp / 2;
      }
      break;
    }
  case PCM_FORMAT_U16_LE:
    {
      uint16_t*		output = buf;
      const uint16_t*	input = data;

      for(i=0; i<samples; i++){
	uint32_t tmp;
	uint16_t stmp;

	tmp = UINT16_FROM_LE(*input);
	input++;
	tmp += UINT16_FROM_LE(*input);
	input++;
	stmp = tmp / 2;
	*output++ = UINT16_TO_LE(stmp);
      }
      break;
    }
  case PCM_FORMAT_U16_BE:
    {
      uint16_t*		output = buf;
      const uint16_t*	input = data;

      for(i=0; i<samples; i++){
	uint32_t tmp;
	uint16_t stmp;

	tmp = UINT16_FROM_BE(*input);
	input++;
	tmp += UINT16_FROM_BE(*input);
	input++;
	stmp = tmp / 2;
	*output++ = UINT16_TO_BE(stmp);
      }
      break;
    }
  case PCM_FORMAT_S16_LE:
    {
      int16_t*		output = buf;
      const int16_t*	input = data;

      for(i=0; i<samples; i++){
	int32_t tmp;
	int16_t stmp;

	tmp = INT16_FROM_LE(*input);
	input++;
	tmp += INT16_FROM_LE(*input);
	input++;
	stmp = tmp / 2;
	*output++ = INT16_TO_LE(stmp);
      }
      break;
    }
  case PCM_FORMAT_S16_BE:
    {
      int16_t*		output = buf;
      const int16_t*	input = data;

      for(i=0; i<samples; i++){
	int32_t tmp;
	int16_t stmp;

	tmp = INT16_FROM_BE(*input);
	input++;
	tmp += INT16_FROM_BE(*input);
	input++;
	stmp = tmp / 2;
	*output++ = INT16_TO_BE(stmp);
      }
      break;
    }
  case PCM_FORMAT_S24_LE:
  case PCM_FORMAT_S24_BE:
  case PCM_FORMAT_U24_LE:
  case PCM_FORMAT_U24_BE:
  case PCM_FORMAT_S32_LE:
  case PCM_FORMAT_S32_BE:
  case PCM_FORMAT_U32_LE:
  case PCM_FORMAT_U32_BE:
    {
      /* TODO */
      fprintf(stderr, "map_2_to_1: unsupported format\n");
      return(0);
    }
  default:
    fprintf(stderr, "map_2_to_1: unknown format\n");
    return(0);
  }
  return(out_len);
}

static size_t map_2_to_4(void* buf, const size_t buf_len, const void* data, const size_t data_len, lamipPCMFormat fmt)
{
  int		samples;
  int		samplesize;
  size_t	out_len;

  lamip_pcm_formatDecomp(fmt, &samplesize, 0, 0);
  samples = data_len / (2*samplesize);
  out_len = samples* 4 *samplesize;

  switch(fmt){
  case PCM_FORMAT_S8:
  case PCM_FORMAT_U8:
    {
      int8_t*		output = buf;
      const int8_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	input += 2;
      }
      break;
    }
  case PCM_FORMAT_S16_LE:
  case PCM_FORMAT_S16_BE:
  case PCM_FORMAT_U16_LE:
  case PCM_FORMAT_U16_BE:
    {
      uint16_t*		output = buf;
      const uint16_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	input += 2;
      }
      break;
    }
  case PCM_FORMAT_S24_LE:
  case PCM_FORMAT_S24_BE:
  case PCM_FORMAT_U24_LE:
  case PCM_FORMAT_U24_BE:
    {
      uint32_t*		output = buf;
      const uint32_t*	input = data;

      fprintf(stderr, "map_2_to_4 - 24bits format set to 32 (TODO)\n");
      while(samples--){
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	input += 2;
      }
      break;
    }
  case PCM_FORMAT_S32_LE:
  case PCM_FORMAT_S32_BE:
  case PCM_FORMAT_U32_LE:
  case PCM_FORMAT_U32_BE:
    {
      uint32_t*		output = buf;
      const uint32_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	input += 2;
      }
      break;
    }
  default:
    {
      fprintf(stderr, "map_2_to_4: unknown format\n");
      return(0);
      break;
    }
  }
  return(out_len);
}

static size_t map_2_to_6(void *buf, const size_t buf_len, const void *data, const size_t data_len, lamipPCMFormat fmt)
{
  int samples, samplesize;
  size_t out_len;

  lamip_pcm_formatDecomp(fmt, &samplesize, 0, 0);
  samples = MIN(data_len / (2 * samplesize), buf_len / (6 * samplesize));
  out_len = samples * (6 * samplesize);

  switch(fmt){
  case PCM_FORMAT_S8:
  case PCM_FORMAT_U8:
    {
      uint8_t*	output = buf;
      const uint8_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	input += 2;
      }
      break;
    }
  case PCM_FORMAT_S16_LE:
  case PCM_FORMAT_S16_BE:
  case PCM_FORMAT_U16_LE:
  case PCM_FORMAT_U16_BE:
    {
      uint16_t*	output = buf;
      const uint16_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	input += 2;
      }
      break;
    }
  case PCM_FORMAT_S24_LE:
  case PCM_FORMAT_S24_BE:
  case PCM_FORMAT_U24_LE:
  case PCM_FORMAT_U24_BE:
    {
      uint32_t*		output = buf;
      const uint32_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	input += 2;
      }
      break;
    }
  case PCM_FORMAT_S32_LE:
  case PCM_FORMAT_S32_BE:
  case PCM_FORMAT_U32_LE:
  case PCM_FORMAT_U32_BE:
    {
      uint32_t*		output = buf;
      const int32_t*	input = data;

      while(samples--){
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	*output++ = *input;
	*output++ = *(input + 1);
	input += 2;
      }
      break;
    }
  default:
    {
      fprintf(stderr, "map_2_to_6: unknown format\n");
      return(0);
      break;
    }
  }
  return(out_len);
}

static size_t map_6_to_2(void *buf, const size_t buf_len, const void *data, const size_t data_len, lamipPCMFormat fmt)
{
  int i, samples, samplesize;
  size_t out_len;

  lamip_pcm_formatDecomp(fmt, &samplesize, NULL, NULL);
  samples = data_len / (6*samplesize);
  // i think this is not so good
  //MIN(data_len / (6 * samplesize), buf_len / (2 * samplesize));
  out_len = samples * (2 * samplesize);

  switch(fmt){
  case PCM_FORMAT_U8:
    {
      uint8_t*		output = buf;
      const uint8_t*	input = data;

      for(i=0; i<samples; i++){
	uint16_t tmp;

	tmp = input[0 + 6 * i] + input[2 + 6 * i] + input[4 + 6 * i];
	output[0 + 2 * i] = tmp / 3;
	tmp = input[1 + 6 * i] + input[3 + 6 * i] + input[5 + 6 * i];
	output[1 + 2 * i] = tmp / 3;
      }
      break;
    }
  case PCM_FORMAT_S8:
    {
      int8_t*		output = buf;
      const int8_t*	input = data;

      for(i=0; i<samples; i++){
	int16_t tmp;

	tmp = input[0 + 6 * i] + input[2 + 6 * i] + input[4 + 6 * i];
	output[0 + 2 * i] = tmp / 3;
	tmp = input[1 + 6 * i] + input[3 + 6 * i] + input[5 + 6 * i];
	output[1 + 2 * i] = tmp / 3;
      }
      break;
    }
  case PCM_FORMAT_U16_LE:
    {
      uint16_t*		output = buf;
      const uint16_t*	input = data;

      for(i=0; i<samples; i++){
	uint16_t stmp;

	stmp = UINT16_FROM_LE(input[0 + 6 * i])
	  | UINT16_FROM_LE(input[2 + 6 * i])
	  | UINT16_FROM_LE(input[4 + 6 * i]);
	output[0 + 2 * i] = UINT16_TO_LE(stmp);

	stmp = UINT16_FROM_LE(input[1 + 6 * i])
	  | UINT16_FROM_LE(input[3 + 6 * i])
	  | UINT16_FROM_LE(input[5 + 6 * i]);
	output[1 + 2 * i] = UINT16_TO_LE(stmp);
      }
      break;
    }
  case PCM_FORMAT_U16_BE:
    {
      uint16_t*		output = buf;
      const uint16_t*	input = data;

      for(i=0; i<samples; i++){
	uint16_t stmp;

	stmp = UINT16_FROM_BE(input[0 + 6 * i])
	  | UINT16_FROM_BE(input[2 + 6 * i])
	  | UINT16_FROM_BE(input[4 + 6 * i]);
	output[0 + 2 * i] = UINT16_TO_BE(stmp);

	stmp = UINT16_FROM_BE(input[1 + 6 * i])
	  | UINT16_FROM_BE(input[3 + 6 * i])
	  | UINT16_FROM_BE(input[5 + 6 * i]);
	output[1 + 2 * i] = UINT16_TO_BE(stmp);
      }
      break;
    }
  case PCM_FORMAT_S16_LE:
    {
      int16_t*		output = buf;
      const int16_t*	input = data;

      for(i=0; i<samples; i++){
	int32_t tmp;
	//int16_t stmp;

	tmp =INT16_FROM_LE(input[0 + 6 * i])
		| INT16_FROM_LE(input[2 + 6 * i])
		| INT16_FROM_LE(input[4 + 6 * i]);
	//stmp = tmp / 3;
	output[0 + 2 * i] = INT16_TO_LE(tmp);

	tmp =INT16_FROM_LE(input[1 + 6 * i])
		| INT16_FROM_LE(input[3 + 6 * i])
		| INT16_FROM_LE(input[5 + 6 * i]);
	//stmp = tmp / 3;
	output[1 + 2 * i] = INT16_TO_LE(tmp);
      }
      break;
    }
  case PCM_FORMAT_S16_BE:
    {
      int16_t*		output = buf;
      const int16_t*	input = data;

      for(i=0; i<samples; i++){
	int32_t tmp;
	int16_t stmp;

	tmp =INT16_FROM_BE(input[0 + 6 * i])
		+ INT16_FROM_BE(input[2 + 6 * i])
		+ INT16_FROM_BE(input[4 + 6 * i]);
	stmp = tmp / 3;
	output[0 + 2 * i] = INT16_TO_BE(stmp);

	tmp =INT16_FROM_BE(input[1 + 6 * i])
		+ INT16_FROM_BE(input[3 + 6 * i])
		+ INT16_FROM_BE(input[5 + 6 * i]);
	stmp = tmp / 3;
	output[1 + 2 * i] = INT16_TO_BE(stmp);
      }
      break;
    }
  case PCM_FORMAT_S24_LE:
  case PCM_FORMAT_S24_BE:
  case PCM_FORMAT_U24_LE:
  case PCM_FORMAT_U24_BE:
  case PCM_FORMAT_S32_LE:
  case PCM_FORMAT_S32_BE:
  case PCM_FORMAT_U32_LE:
  case PCM_FORMAT_U32_BE:
    {
      /* TODO */
      fprintf(stderr, "map_6_to_2: unsupported format\n");
      return(0);
    }
  default:
    {
      fprintf(stderr, "map_6_to_2: unknown format\n");
      return(0);
    }
  }
  return(out_len);
}

static size_t map_6_to_4(void *buf, const size_t buf_len, const void *data, const size_t data_len, lamipPCMFormat fmt)
{
  int		i;
  int		samples;
  int		samplesize;
  size_t	out_len;

  lamip_pcm_formatDecomp(fmt, &samplesize, 0, 0);
  samples = data_len / (6*samplesize);
  out_len = samples* 4 *samplesize;

  switch(fmt){
  case PCM_FORMAT_U8:
    {
      uint8_t*		output = buf;
      const uint8_t*	input = data;

      for(i=0; i<samples; i++){
	output[4*i+0] = input[6*i+0] | input[6*i+4];
	output[4*i+1] = input[6*i+1] | input[6*i+5];
	output[4*i+2] = input[6*i+2] | input[6*i+4];
	output[4*i+3] = input[6*i+3] | input[6*i+5];
      }
      break;
    }
  case PCM_FORMAT_S8:
    {
      int8_t*		output = buf;
      const int8_t*	input = data;

      for(i=0; i<samples; i++){
	output[4*i+0] = input[6*i+0] | input[6*i+4];
	output[4*i+1] = input[6*i+1] | input[6*i+5];
	output[4*i+2] = input[6*i+2] | input[6*i+4];
	output[4*i+3] = input[6*i+3] | input[6*i+5];
      }
      break;
    }
  case PCM_FORMAT_U16_LE:
  case PCM_FORMAT_U16_BE:
    {
      uint16_t*		output = buf;
      const uint16_t*	input = data;

      for(i=0; i<samples; i++){
	output[4*i+0] = input[6*i+0] | input[6*i+4];
	output[4*i+1] = input[6*i+1] | input[6*i+5];
	output[4*i+2] = input[6*i+2] | input[6*i+4];
	output[4*i+3] = input[6*i+3] | input[6*i+5];
      }
      break;
    }
  case PCM_FORMAT_S16_LE:
    {
      int16_t*		output = buf;
      const int16_t*	input = data;

      for(i=0; i<samples; i++){
	output[4*i+0] = input[6*i+0] | input[6*i+4];
	output[4*i+1] = input[6*i+1] | input[6*i+5];
	output[4*i+2] = input[6*i+2] | input[6*i+4];
	output[4*i+3] = input[6*i+3] | input[6*i+5];
      }
      break;
    }
  case PCM_FORMAT_S16_BE:
    {
      int16_t*	output = buf;
      const int16_t*	input = data;

      for(i=0; i<samples; i++){
	output[4*i+0] = input[6*i+0] | input[6*i+4];
	output[4*i+1] = input[6*i+1] | input[6*i+5];
	output[4*i+2] = input[6*i+2] | input[6*i+4];
	output[4*i+3] = input[6*i+3] | input[6*i+5];
      }
      break;
    }
  case PCM_FORMAT_S24_LE:
  case PCM_FORMAT_S24_BE:
  case PCM_FORMAT_U24_LE:
  case PCM_FORMAT_U24_BE:
  case PCM_FORMAT_S32_LE:
  case PCM_FORMAT_S32_BE:
  case PCM_FORMAT_U32_LE:
  case PCM_FORMAT_U32_BE:
    {
      /* TODO */
      fprintf(stderr, "map_6_to_4: 24/32 bits (TODO)\n");
      return(0);
    }
  default:
    {
      fprintf(stderr, "map_6_to_4: unknown format\n");
      return(0);
    }
  }
  return(out_len);
}
